#include <iostream>
#include <vector>
#include <thread>
#include <mutex>

using namespace std;

void incrementCounterUsingMutex(int *pCounter, mutex &counterMutex) {
    for (int i = 0; i < 10000000; i++) {
        counterMutex.lock();
        (*pCounter)++;
        counterMutex.unlock();
    }
}

void incrementCounterUsingLockGuard(int *pCounter, mutex &counterMutex) {
    for (int i = 0; i < 10000000; i++) {
        // Performance check: what if the locking is done outside of the for?
        lock_guard<mutex> guard(counterMutex);  // The counterMutex is locked.
        (*pCounter)++;
    } // <-- guard goes out of scope, counterMutex is unlocked in the destructor of guard
}

void incrementCounterUsingUniqueLock1(int *pCounter, mutex &counterMutex) {
    for (int i = 0; i < 10000000; i++) {
        unique_lock<mutex> lock(counterMutex); // The counterMutex is locked.
        (*pCounter)++;
    } // <-- lock goes out of scope, counterMutex is unlocked in the destructor of lock
}

void incrementCounterUsingUniqueLock2(int *pCounter, mutex &counterMutex) {
    for (int i = 0; i < 10000000; i++) {
        unique_lock<mutex> lock(counterMutex, defer_lock);  // The counterMutex is not locked yet!
        lock.lock();
        (*pCounter)++;
    } // <-- lock goes out of scope, counterMutex is unlocked in the destructor of lock
}

int main() {
    int counter = 0;
    mutex counterMutex;

    vector<thread> workers;

    /*
    // incrementCounterUsingMutex
    for (int workerIndex = 0; workerIndex < 4; workerIndex++) {
        workers.push_back(thread(incrementCounterUsingMutex, &counter, ref(counterMutex)));
    }
     */

    /*
    // incrementCounterUsingLockGuard
    for (int workerIndex = 0; workerIndex < 4; workerIndex++) {
        workers.push_back(thread(incrementCounterUsingLockGuard, &counter, ref(counterMutex)));
    }
    */

    /*
    // incrementCounterUsingUniqueLock1
    for (int workerIndex = 0; workerIndex < 4; workerIndex++) {
        workers.push_back(thread(incrementCounterUsingUniqueLock1, &counter, ref(counterMutex)));
    }
    */

    // incrementCounterUsingUniqueLock2
    for (int workerIndex = 0; workerIndex < 4; workerIndex++) {
        workers.push_back(thread(incrementCounterUsingUniqueLock2, &counter, ref(counterMutex)));
    }

    for (auto &worker : workers) {
        worker.join();
    }

    cout << "Value of the counter variable: " << counter << endl;

    return 0;
}

